@@ -1,35 +1,8 @@
 #!/usr/bin/env python3
 #
-#  Copyright (c) 2013-2021, Intel Corporation
-#  All rights reserved.
+#  Copyright (c) 2013-2024, Intel Corporation
 #
-#  Redistribution and use in source and binary forms, with or without
-#  modification, are permitted provided that the following conditions are
-#  met:
-#
-#    * Redistributions of source code must retain the above copyright
-#      notice, this list of conditions and the following disclaimer.
-#
-#    * Redistributions in binary form must reproduce the above copyright
-#      notice, this list of conditions and the following disclaimer in the
-#      documentation and/or other materials provided with the distribution.
-#
-#    * Neither the name of Intel Corporation nor the names of its
-#      contributors may be used to endorse or promote products derived from
-#      this software without specific prior written permission.
-#
-#
-#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
-#   IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
-#   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
-#   PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
-#   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
-#   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
-#   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
-#   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
-#   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
-#   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
-#   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+#  SPDX-License-Identifier: BSD-3-Clause
 
 # Supported operating systems
 from enum import Enum, unique
@@ -110,18 +83,27 @@ class Host(object):
 # The description of testing target configuration
 class TargetConfig(object):
     def __init__(self, arch, target, cpu):
-        self.arch = arch
+        if arch == "x86_64" or arch == "x86-64":
+            self.arch = "x86-64"
+        else:
+            self.arch = arch
         self.target = target
-        self.genx = target.find("genx") != -1
+        self.xe = target.find("gen9") != -1 or target.find("xe") != -1
         self.set_cpu(cpu)
         self.set_target()
 
-    def is_genx(self):
-        return self.genx
+    def is_xe(self):
+        return self.xe
 
     def set_cpu(self, cpu):
         if cpu is not None:
             self.cpu = cpu
+            # Alias all of acm-* devices to dg2.
+            if cpu.startswith("acm-"):
+                self.cpu = "dg2"
+            # Alias all of mtl-* devices to mtl.
+            if cpu.startswith("mtl-"):
+                self.cpu = "mtl"
         else:
             self.cpu = "unspec"
 
@@ -204,38 +186,6 @@ def check_print_output(output):
     else:
         return lines[0:len(lines)//2] == lines[len(lines)//2:len(lines)]
 
-def prerun_debug_check_genx():
-    os.environ['IGC_DumpToCurrentDir'] = '1'
-    os.environ['IGC_ShaderDumpEnable'] = '1'
-    os.environ['ISPCRT_IGC_OPTIONS'] = '+ -g'
-
-def postrun_debug_check_genx(test_name, exe_wd):
-    objdumpProc = subprocess.run('/usr/bin/objdump -h *_dwarf.elf', shell=True, cwd=exe_wd, universal_newlines=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
-    readelfProc = subprocess.run('/usr/bin/readelf --debug-dump *_dwarf.elf', shell=True, cwd=exe_wd, universal_newlines=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
-
-    # Cleanup files used to check debug information
-    # since this directory may be reused by other tests.
-    subprocess.run('rm -rf *_dwarf.elf', shell=True)
-
-    if objdumpProc.returncode != 0:
-        print("[DEBUG CHECK] " + test_name + " - objdump failed - return code: " + str(objdumpProc.returncode))
-        return False
-    elif objdumpProc.stdout.find('.debug') == -1:
-        print("[DEBUG CHECK] " + test_name + " - objdump failed - couldn't find .debug info in *_dwarf.elf file")
-        return False
-    elif readelfProc.returncode != 0:
-        print("[DEBUG CHECK] " + test_name + " - readelf failed - return code: " + str(objdumpProc.returncode))
-        return False
-    elif readelfProc.stdout.find('Warning:') != -1 or readelfProc.stdout.find('Error:') != -1:
-        readelfOutput = readelfProc.stdout.splitlines()
-        print("[DEBUG CHECK] " + test_name + " - readelf return Warning / Error: \n")
-        for line in readelfOutput:
-            if line.find('Warning:') != -1 or line.find('Error:') != -1:
-                print(line)
-        return False
-    return True
-
-
 # run the commands in cmd_list
 def run_cmds(compile_cmds, run_cmd, filename, expect_failure, sig, exe_wd="."):
     for cmd in compile_cmds:
@@ -248,9 +198,6 @@ def run_cmds(compile_cmds, run_cmd, filename, expect_failure, sig, exe_wd="."):
             return Status.Compfail
 
     if not options.save_bin:
-        if options.debug_check and options.ispc_output == "spv":
-            prerun_debug_check_genx()
-
         (return_code, output, timeout) = run_command(run_cmd, options.test_time, cwd=exe_wd)
         if sig < 32:
             run_failed = (return_code != 0) or timeout
@@ -261,9 +208,6 @@ def run_cmds(compile_cmds, run_cmd, filename, expect_failure, sig, exe_wd="."):
                 print_debug("Print outputs check failed\n", s, run_tests_log)
             run_failed = (return_code != 0) or not output_equality or timeout
 
-        if options.debug_check and options.ispc_output == "spv":
-            run_failed = not postrun_debug_check_genx(run_cmd, exe_wd)
-
     else:
         run_failed = 0
 
@@ -286,9 +230,9 @@ def add_prefix(path, host, target):
     # On Windows we run tests in tmp dir, so the root is one level up.
         input_prefix = "..\\"
     else:
-        # For Gen target we run tests in tmp dir since output file has
+        # For Xe target we run tests in tmp dir since output file has
         # the same name for all tests, so the root is one level up
-        if target.is_genx():
+        if target.is_xe():
             input_prefix = "../"
         else:
             input_prefix = ""
@@ -317,10 +261,9 @@ def add_prefix(path, host, target):
 #
 # Examples:
 #
-# 1. Run only on arch genx32 or arch genx64:
+# 1. Run only on arch xe64:
 # // rule: skip on arch=*
-# // rule: run on arch=genx32
-# // rule: run on arch=genx64
+# // rule: run on arch=xe64
 #
 # 2. Run only on Linux OS:
 # // rule: skip on OS=*
@@ -341,7 +284,9 @@ def check_if_skip_test(filename, host, target):
     else:
         oss = "unknown"
 
-    rule_values = {"arch": target.arch, "OS": oss, "cpu": target.cpu}
+    target_cpu = target.cpu
+
+    rule_values = {"arch": target.arch, "OS": oss, "cpu": target_cpu}
 
     test_file_path = add_prefix(filename, host, target)
     with open(test_file_path) as test_file:
@@ -376,8 +321,8 @@ def run_test(testname, host, target):
     # ispc_exe_rel is a relative path to ispc
     filename = add_prefix(testname, host, target)
 
-    # Debug check is now supported only for genx
-    if options.debug_check and target.is_genx():
+    # Debug check is now supported only for xe
+    if options.debug_check and target.is_xe():
         ispc_exe_rel = add_prefix(host.ispc_cmd + " -g", host, target)
     else:
         ispc_exe_rel = add_prefix(host.ispc_cmd, host, target)
@@ -421,7 +366,7 @@ def run_test(testname, host, target):
         match = -1
         for line in file:
             # look for lines with 'export'...
-            if line.find("export") == -1:
+            if line.find("task") == -1 and line.find("export") == -1:
                 continue
             # one of them should have a function with one of the
             # declarations in sig2def
@@ -433,96 +378,86 @@ def run_test(testname, host, target):
 
         # Figure out target width
         width = -1
-        target_match = re.match('(.*-i[0-9]*x)([0-9]*)', options.target)
+        target_match = re.match('.*-(i[0-9]*)?x([0-9]*)', options.target)
         # If target does not contain width in a standard way:
         if target_match == None:
-            if re.match('(sse[0-9]*-x2)', options.target) != None:
-                width = 8
-            elif re.match('(sse[0-9]*)', options.target) != None:
-                width = 4
-            elif re.match('(avx[0-9]*-x2)', options.target) != None:
-                width = 16
-            elif re.match('(avx[0-9]*)', options.target) != None:
-                width = 8
-            elif options.target == "neon":
-                width = 4
-            elif re.search('genx', options.target) != None:
-                width = 16
-                target_match = re.match('(genx-x)([0-9]*)', options.target)
-                if target_match != None:
-                    width = int(target_match.group(2))
-                else:
-                    width = 16
-        else:
-            width = int(target_match.group(2))
-        if width == -1:
-            error("unable to determine target width for target %s\n" % options.target, 0)
+            error("Unable to detect the target width for target %s\nOnly canonical form of the target names is supported, deprecated forms are not supported" % options.target, 0)
             return Status.Compfail
+        width = int(target_match.group(2))
+
         if match == -1:
-            error("unable to find function signature in test %s\n" % testname, 0)
+            error("Unable to find function signature in test %s\n" % testname, 0)
             return Status.Compfail
         else:
-            genx_target = options.target
+            xe_target = options.target
             if host.is_windows():
-                if target.is_genx():
-                    obj_name = "test_genx.bin" if options.ispc_output == "ze" else "test_genx.spv"
+                if target.is_xe():
+                    obj_name = "test_xe.bin" if options.ispc_output == "ze" else "test_xe.spv"
                 else:
                     obj_name = "%s.obj" % os.path.basename(filename)
 
-                if target.arch == "wasm32":
+                if target.arch == "wasm32" or target.arch == "wasm64":
                     exe_name = "%s.js" % os.path.realpath(filename)
                 else:
                     exe_name = "%s.exe" % os.path.basename(filename)
 
                 cc_cmd = "%s /I. /Zi /nologo /DTEST_SIG=%d /DTEST_WIDTH=%d %s %s /Fe%s" % \
                          (options.compiler_exe, match, width, add_prefix("test_static.cpp", host, target), obj_name, exe_name)
-                if target.is_genx():
+                if target.is_xe():
                     cc_cmd = "%s /I. /I%s\\include /nologo /DTEST_SIG=%d /DTEST_WIDTH=%d %s %s /Fe%s ze_loader.lib /link /LIBPATH:%s\\lib" % \
                          (options.compiler_exe, options.l0loader, match, width, " /DTEST_ZEBIN" if options.ispc_output == "ze" else " /DTEST_SPV", \
                          add_prefix("test_static_l0.cpp", host, target), exe_name, options.l0loader)
                 if should_fail:
                     cc_cmd += " /DEXPECT_FAILURE"
             else:
-                if target.is_genx():
-                    obj_name = "test_genx.bin" if options.ispc_output == "ze" else "test_genx.spv"
+                if target.is_xe():
+                    obj_name = "test_xe.bin" if options.ispc_output == "ze" else "test_xe.spv"
                 else:
                     obj_name = "%s.o" % testname
 
-                if target.arch == "wasm32":
+                if target.arch == "wasm32" or target.arch == "wasm64":
                     exe_name = "%s.js" % os.path.realpath(testname)
                 else:
                     exe_name = "%s.run" % testname
 
                 if target.arch == 'arm':
                     gcc_arch = '--with-fpu=hardfp -marm -mfpu=neon -mfloat-abi=hard'
-                elif target.arch == 'x86' or target.arch == "wasm32" or target.arch == 'genx32':
+                elif target.arch == 'x86' or target.arch == "wasm32":
                     gcc_arch = '-m32'
                 elif target.arch == 'aarch64':
                     gcc_arch = '-march=armv8-a'
+                elif target.arch == 'wasm64':
+                    gcc_arch = '-sMEMORY64'
                 else:
                     gcc_arch = '-m64'
 
                 cc_cmd = "%s -O2 -I. %s test_static.cpp -DTEST_SIG=%d -DTEST_WIDTH=%d %s -o %s" % \
                     (options.compiler_exe, gcc_arch, match, width, obj_name, exe_name)
 
+                # Produce position independent code for both c++ and ispc compilations.
+                # The motivation for this is that Clang 15 changed default
+                # from "-mrelocation-model static" to "-mrelocation-model pic", so
+                # we enable PIC compilation to have it consistently regardless compiler version.
+                cc_cmd += ' -fPIE'
                 if should_fail:
                     cc_cmd += " -DEXPECT_FAILURE"
 
-                if target.is_genx():
+                if target.is_xe():
                     exe_name = "%s.run" % os.path.basename(testname)
-                    cc_cmd = "%s -O2 -I. -lze_loader \
+                    cc_cmd = "%s -O0 -I. -I %s/include -lze_loader -L %s/lib \
                             %s %s -DTEST_SIG=%d -DTEST_WIDTH=%d -o %s" % \
-                            (options.compiler_exe, gcc_arch, add_prefix("test_static_l0.cpp", host, target), match, width, exe_name)
+                            (options.compiler_exe, options.l0loader, options.l0loader, gcc_arch, add_prefix("test_static_l0.cpp", host, target),
+                             match, width, exe_name)
                     exe_name = "./" + exe_name
                     cc_cmd += " -DTEST_ZEBIN" if options.ispc_output == "ze" else " -DTEST_SPV"
+            ispc_cmd = ispc_exe_rel + " --pic --woff %s -o %s --arch=%s --target=%s -DTEST_SIG=%d" % \
+                        (filename, obj_name, options.arch, xe_target if target.is_xe() else options.target, match)
 
-            ispc_cmd = ispc_exe_rel + " --woff %s -o %s --arch=%s --target=%s" % \
-                        (filename, obj_name, options.arch, genx_target if target.is_genx() else options.target)
-
-            if target.is_genx():
+            if target.is_xe():
                 ispc_cmd += " --emit-zebin" if options.ispc_output == "ze" else " --emit-spirv"
-            if options.cpu != None:
-                ispc_cmd += " --cpu="+ options.cpu
+                ispc_cmd += " -DISPC_GPU"
+            if options.device != None:
+                ispc_cmd += " --device="+ options.device
 
             if options.opt == 'O0':
                 ispc_cmd += " -O0"
@@ -532,9 +467,11 @@ def run_test(testname, host, target):
                 ispc_cmd += " -O2"
 
         exe_wd = "."
-        if target.arch == "wasm32":
+        if target.arch == "wasm32" or target.arch == "wasm64":
             cc_cmd += " -D__WASM__"
-            options.wrapexe = "v8 --experimental-wasm-simd"
+            options.wrapexe = os.environ["EMSDK_NODE"]
+            if target.arch == "wasm64":
+                options.wrapexe += " --experimental-wasm-memory64"
             exe_wd = os.path.realpath("./tests")
         # compile the ispc code, make the executable, and run it...
         ispc_cmd += " -h " + filename + ".h"
@@ -575,7 +512,7 @@ def run_tasks_from_queue(queue, queue_ret, total_tests_arg, max_test_length_arg,
     global run_tests_log
     run_tests_log = glob_var[4]
 
-    if host.is_windows() or target.is_genx():
+    if host.is_windows() or target.is_xe():
         tmpdir = "tmp%d" % os.getpid()
         while os.access(tmpdir, os.F_OK):
             tmpdir = "%sx" % tmpdir
@@ -617,7 +554,7 @@ def run_tasks_from_queue(queue, queue_ret, total_tests_arg, max_test_length_arg,
         except:
             None
     else:
-        if target.is_genx():
+        if target.is_xe():
             try:
                 os.chdir("..")
                 os.rmdir(tmpdir)
@@ -662,7 +599,7 @@ def file_check(results, host, target):
     ispc_output = options.ispc_output
 # Detect LLVM version
     temp1 = common.take_lines(host.ispc_exe + " --version", "first")
-    temp2 = re.search('LLVM [0-9]*\.[0-9]*', temp1)
+    temp2 = re.search(r'LLVM [0-9]*\.[0-9]*', temp1)
     if temp2 != None:
         llvm_version = temp2.group()
     else:
@@ -670,11 +607,11 @@ def file_check(results, host, target):
 # Detect compiler version
     if OS != "Windows":
         temp1 = common.take_lines(options.compiler_exe + " --version", "first")
-        temp2 = re.search("[0-9]*\.[0-9]*\.[0-9]", temp1)
+        temp2 = re.search(r"[0-9]*\.[0-9]*\.[0-9]", temp1)
         if temp2 == None:
-            temp3 = re.search("[0-9]*\.[0-9]*", temp1)
+            temp3 = re.search(r"[0-9]*\.[0-9]*", temp1)
         else:
-            temp3 = re.search("[0-9]*\.[0-9]*", temp2.group())
+            temp3 = re.search(r"[0-9]*\.[0-9]*", temp2.group())
         compiler_version = options.compiler_exe + temp3.group()
     else:
         compiler_version = "cl"
@@ -767,9 +704,10 @@ def verify():
              ["Linux","Windows","Mac"],["LLVM 3.2","LLVM 3.3","LLVM 3.4","LLVM 3.5","LLVM 3.6","LLVM trunk"],
              ["sse2-i32x4", "sse2-i32x8",
               "sse4-i32x4", "sse4-i32x8", "sse4-i16x8", "sse4-i8x16",
+              "sse4.1-i32x4", "sse4.1-i32x8", "sse4.1-i16x8", "sse4.1-i8x16",
               "avx1-i32x4", "avx1-i32x8", "avx1-i32x16", "avx1-i64x4",
               "avx2-i32x4", "avx2-i32x8", "avx2-i32x16", "avx2-i64x4",
-              "avx512knl-i32x16", "avx512skx-i32x16", "avx512skx-i32x8", "avx512skx-i8x64", "avx512skx-i16x32"]]
+              "avx512knl-x16", "avx512skx-x16", "avx512skx-x8", "avx512skx-x4", "avx512skx-x64", "avx512skx-x32"]]
     for i in range (0,len(f_lines)):
         if f_lines[i][0] == "%":
             continue
@@ -806,7 +744,7 @@ def populate_ex_state(options, target, total_tests, test_result):
 # set compiler exe depending on the OS
 def set_compiler_exe(host, options):
     if options.compiler_exe == None:
-        if options.arch == "wasm32":
+        if options.arch == "wasm32" or options.arch == "wasm64":
           options.compiler_exe = "emcc"
         elif host.is_windows():
             options.compiler_exe = "cl.exe"
@@ -818,12 +756,12 @@ def set_compiler_exe(host, options):
 # set ISPC output format
 def set_ispc_output(target, options):
     if options.ispc_output == None:
-        if target.is_genx():
+        if target.is_xe():
             options.ispc_output = "spv"
         else:
             options.ispc_output = "obj"
     else:
-        if not target.is_genx() and not options.ispc_output=="obj" or target.is_genx() and options.ispc_output=="obj":
+        if not target.is_xe() and not options.ispc_output=="obj" or target.is_xe() and options.ispc_output=="obj":
             error("unsupported test output \"%s\" is specified for target: %s \n" % (options.ispc_output, target.target), 1)
 
 # returns the list of test files
@@ -898,10 +836,10 @@ def run_tests(options1, args, print_version):
     host = Host(platform.system())
     host.set_ispc_cmd(options.ispc_flags)
 
-    target = TargetConfig(options.arch, options.target, options.cpu)
+    target = TargetConfig(options.arch, options.target, options.device)
 
-    if options.debug_check and (not target.is_genx() or not host.is_linux()):
-        print("--debug_check is supported only for genx target and only on Linux OS")
+    if options.debug_check and (not target.is_xe() or not host.is_linux()):
+        print("--debug_check is supported only for xe target and only on Linux OS")
         exit_code = 1
         return 0
 
@@ -1052,11 +990,12 @@ elif platform.machine() == "aarch64":
 elif platform.machine() == "arm64":
     default_target = "neon-i32x4"
     default_arch = "aarch64"
-elif "86" in platform.machine():
+elif "86" in platform.machine() or platform.machine() == "AMD64":
     # Some variant of x86: x86_64, i386, i486, i586, i686
+    # Windows reports platform as AMD64
     pass
 else:
-    print_debug("WARNING: host machine was not recognized", False, "")
+    print_debug("WARNING: host machine was not recognized - " + str(platform.machine()), False, "")
 
 if __name__ == "__main__":
     parser = OptionParser()
@@ -1065,13 +1004,13 @@ if __name__ == "__main__":
     parser.add_option("-f", "--ispc-flags", dest="ispc_flags", help="Additional flags for ispc (-g, -O1, ...)",
                   default="")
     parser.add_option('-t', '--target', dest='target',
-                  help=('Set compilation target. For example: sse4-i32x4, avx2-i32x8, avx512skx-i32x16, etc.'), default=default_target)
+                  help=('Set compilation target. For example: sse4-i32x4, avx2-i32x8, avx512skx-x16, etc.'), default=default_target)
     parser.add_option('-a', '--arch', dest='arch',
-                  help='Set architecture (arm, aarch64, x86, x86-64, genx32, genx64)', default=default_arch)
+                  help='Set architecture (arm, aarch64, x86, x86-64, xe64)', default=default_arch)
     parser.add_option("-c", "--compiler", dest="compiler_exe", help="C/C++ compiler binary to use to run tests",
                   default=None)
     parser.add_option('-o', '--opt', dest='opt', choices=['', 'O0', 'O1', 'O2'], help='Set optimization level passed to the compiler (O0, O1, O2).',
-                  default='')
+                  default='O2')
     parser.add_option('-j', '--jobs', dest='num_jobs', help='Maximum number of jobs to run in parallel',
                   default="1024", type="int")
     parser.add_option('-v', '--verbose', dest='verbose', help='Enable verbose output',
@@ -1088,7 +1027,7 @@ if __name__ == "__main__":
                   action = "store_true")
     parser.add_option("--file", dest='in_file', help='file to save run_tests output', default="")
     parser.add_option("--l0loader", dest='l0loader', help='Path to L0 loader', default="")
-    parser.add_option("--cpu", dest='cpu', help='Specify target ISPC CPU. For example: core2, skx, cortex-a9, SKL, TGLLP, etc.', default=None)
+    parser.add_option("--device", dest='device', help='Specify target ISPC device. For example: core2, skx, cortex-a9, skl, tgllp, acm-g11, etc.', default=None)
     parser.add_option("--ispc_output", dest='ispc_output', choices=['obj', 'spv', 'ze'], help='Specify ISPC output', default=None)
     parser.add_option("--fail_db", dest='fail_db', help='File to use as a fail database', default='fail_db.txt', type=str)
     parser.add_option("--debug_check", dest='debug_check', help='Run tests in debug mode with validating debug info', default=False, action="store_true")
