@@ -2,8 +2,7 @@ WD=`pwd`
 git clone https://github.com/emscripten-core/emsdk.git && \
 cd emsdk
 git pull && \
-git checkout 2.0.4 && \
-./emsdk install sdk-1.39.11-64bit && \
-./emsdk activate sdk-1.39.11-64bit  && \
+./emsdk install 3.1.17 && \
+./emsdk activate 3.1.17  && \
 source ./emsdk_env.sh
 cd $WD
