@@ -1,34 +1,7 @@
 /*
-  Copyright (c) 2010-2021, Intel Corporation
-  All rights reserved.
+  Copyright (c) 2010-2023, Intel Corporation
 
-  Redistribution and use in source and binary forms, with or without
-  modification, are permitted provided that the following conditions are
-  met:
-
-    * Redistributions of source code must retain the above copyright
-      notice, this list of conditions and the following disclaimer.
-
-    * Redistributions in binary form must reproduce the above copyright
-      notice, this list of conditions and the following disclaimer in the
-      documentation and/or other materials provided with the distribution.
-
-    * Neither the name of Intel Corporation nor the names of its
-      contributors may be used to endorse or promote products derived from
-      this software without specific prior written permission.
-
-
-   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
-   IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
-   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
-   PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
-   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
-   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
-   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
-   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
-   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
-   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
-   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+  SPDX-License-Identifier: BSD-3-Clause
 */
 
 /** @file decl.cpp
@@ -75,8 +48,8 @@ static void lPrintTypeQualifiers(int typeQualifiers) {
     the type, returning the type that is the result.
 */
 static const Type *lApplyTypeQualifiers(int typeQualifiers, const Type *type, SourcePos pos) {
-    if (type == NULL)
-        return NULL;
+    if (type == nullptr)
+        return nullptr;
 
     if ((typeQualifiers & TYPEQUAL_CONST) != 0) {
         type = type->GetAsConstType();
@@ -107,7 +80,7 @@ static const Type *lApplyTypeQualifiers(int typeQualifiers, const Type *type, So
                        "qualifiers.");
 
         const Type *unsignedType = type->GetAsUnsignedType();
-        if (unsignedType != NULL)
+        if (unsignedType != nullptr)
             type = unsignedType;
         else {
             const Type *resolvedType = type->ResolveUnboundVariability(Variability::Varying);
@@ -135,7 +108,7 @@ DeclSpecs::DeclSpecs(const Type *t, StorageClass sc, int tq) {
     typeQualifiers = tq;
     soaWidth = 0;
     vectorSize = 0;
-    if (t != NULL) {
+    if (t != nullptr) {
         if (m->symbolTable->ContainsType(t)) {
             // Typedefs might have uniform/varying qualifiers inside.
             if (t->IsVaryingType()) {
@@ -150,17 +123,17 @@ DeclSpecs::DeclSpecs(const Type *t, StorageClass sc, int tq) {
 const Type *DeclSpecs::GetBaseType(SourcePos pos) const {
     const Type *retType = baseType;
 
-    if (retType == NULL) {
+    if (retType == nullptr) {
         Warning(pos, "No type specified in declaration.  Assuming int32.");
         retType = AtomicType::UniformInt32->GetAsUnboundVariabilityType();
     }
 
     if (vectorSize > 0) {
         const AtomicType *atomicType = CastType<AtomicType>(retType);
-        if (atomicType == NULL) {
+        if (atomicType == nullptr) {
             Error(pos, "Only atomic types (int, float, ...) are legal for vector "
                        "types.");
-            return NULL;
+            return nullptr;
         }
         retType = new VectorType(atomicType, vectorSize);
     }
@@ -170,18 +143,18 @@ const Type *DeclSpecs::GetBaseType(SourcePos pos) const {
     if (soaWidth > 0) {
         const StructType *st = CastType<StructType>(retType);
 
-        if (st == NULL) {
+        if (st == nullptr) {
             Error(pos,
                   "Illegal to provide soa<%d> qualifier with non-struct "
                   "type \"%s\".",
                   soaWidth, retType ? retType->GetString().c_str() : "NULL");
-            return NULL;
+            return nullptr;
         } else if (soaWidth <= 0 || (soaWidth & (soaWidth - 1)) != 0) {
             Error(pos,
                   "soa<%d> width illegal. Value must be positive power "
                   "of two.",
                   soaWidth);
-            return NULL;
+            return nullptr;
         }
 
         if (st->IsUniformType()) {
@@ -189,13 +162,13 @@ const Type *DeclSpecs::GetBaseType(SourcePos pos) const {
                   "\"uniform\" qualifier and \"soa<%d>\" qualifier can't "
                   "both be used in a type declaration.",
                   soaWidth);
-            return NULL;
+            return nullptr;
         } else if (st->IsVaryingType()) {
             Error(pos,
                   "\"varying\" qualifier and \"soa<%d>\" qualifier can't "
                   "both be used in a type declaration.",
                   soaWidth);
-            return NULL;
+            return nullptr;
         } else
             retType = st->GetAsSOAType(soaWidth);
 
@@ -218,6 +191,8 @@ static const char *lGetStorageClassName(StorageClass storageClass) {
         return "extern";
     case SC_EXTERN_C:
         return "extern \"C\"";
+    case SC_EXTERN_SYCL:
+        return "extern \"SYCL\"";
     case SC_STATIC:
         return "static";
     case SC_TYPEDEF:
@@ -245,27 +220,31 @@ void DeclSpecs::Print() const {
 // Declarator
 
 Declarator::Declarator(DeclaratorKind dk, SourcePos p) : pos(p), kind(dk) {
-    child = NULL;
+    child = nullptr;
     typeQualifiers = 0;
     storageClass = SC_NONE;
     arraySize = -1;
-    type = NULL;
-    initExpr = NULL;
+    type = nullptr;
+    initExpr = nullptr;
 }
 
 void Declarator::InitFromDeclSpecs(DeclSpecs *ds) {
     const Type *baseType = ds->GetBaseType(pos);
+    if (!baseType) {
+        AssertPos(pos, m->errorCount > 0);
+        return;
+    }
 
     InitFromType(baseType, ds);
 
-    if (type == NULL) {
+    if (type == nullptr) {
         AssertPos(pos, m->errorCount > 0);
         return;
     }
 
     storageClass = ds->storageClass;
 
-    if (ds->declSpecList.size() > 0 && CastType<FunctionType>(type) == NULL) {
+    if (ds->declSpecList.size() > 0 && CastType<FunctionType>(type) == nullptr) {
         Error(pos,
               "__declspec specifiers for non-function type \"%s\" are "
               "not used.",
@@ -273,10 +252,17 @@ void Declarator::InitFromDeclSpecs(DeclSpecs *ds) {
     }
 }
 
-void Declarator::Print(int indent) const {
-    printf("%*cdeclarator: [", indent, ' ');
-    pos.Print();
+void Declarator::Print() const {
+    Indent indent;
+    indent.pushSingle();
+    Print(indent);
+    fflush(stdout);
+}
 
+void Declarator::Print(Indent &indent) const {
+    indent.Print("Declarator", pos);
+
+    printf("[");
     lPrintTypeQualifiers(typeQualifiers);
     printf("%s ", lGetStorageClassName(storageClass));
     if (name.size() > 0)
@@ -307,23 +293,32 @@ void Declarator::Print(int indent) const {
         FATAL("Unhandled declarator kind");
     }
 
-    if (initExpr != NULL) {
-        printf(" = (");
-        initExpr->Print();
-        printf(")");
+    printf("]\n");
+
+    int kids = (initExpr ? 1 : 0) + functionParams.size() + (child ? 1 : 0);
+    indent.pushList(kids);
+
+    if (initExpr != nullptr) {
+        indent.setNextLabel("init");
+        initExpr->Print(indent);
     }
 
     if (functionParams.size() > 0) {
         for (unsigned int i = 0; i < functionParams.size(); ++i) {
-            printf("\n%*cfunc param %d:\n", indent, ' ', i);
-            functionParams[i]->Print(indent + 4);
+            static constexpr std::size_t BUFSIZE{20};
+            char buffer[BUFSIZE];
+            snprintf(buffer, BUFSIZE, "func param %d", i);
+            indent.setNextLabel(buffer);
+            functionParams[i]->Print(indent);
         }
     }
 
-    if (child != NULL)
-        child->Print(indent + 4);
+    if (child != nullptr) {
+        indent.setNextLabel("child");
+        child->Print(indent);
+    }
 
-    printf("]\n");
+    indent.Done();
 }
 
 void Declarator::InitFromType(const Type *baseType, DeclSpecs *ds) {
@@ -361,14 +356,14 @@ void Declarator::InitFromType(const Type *baseType, DeclSpecs *ds) {
         // All of the type qualifiers should be in the DeclSpecs for the
         // base declarator
         AssertPos(pos, typeQualifiers == 0);
-        AssertPos(pos, child == NULL);
+        AssertPos(pos, child == nullptr);
         type = baseType;
     } else if (kind == DK_POINTER) {
         /* For now, any pointer to an SOA type gets the slice property; if
            we add the capability to declare pointers as slices or not,
            we'll want to set this based on a type qualifier here. */
         const Type *ptrType = new PointerType(baseType, variability, isConst, baseType->IsSOAType());
-        if (child != NULL) {
+        if (child != nullptr) {
             child->InitFromType(ptrType, ds);
             type = child->type;
             name = child->name;
@@ -388,13 +383,13 @@ void Declarator::InitFromType(const Type *baseType, DeclSpecs *ds) {
             return;
         }
         // The parser should disallow this already, but double check.
-        if (CastType<ReferenceType>(baseType) != NULL) {
+        if (CastType<ReferenceType>(baseType) != nullptr) {
             Error(pos, "References to references are illegal.");
             return;
         }
 
         const Type *refType = new ReferenceType(baseType);
-        if (child != NULL) {
+        if (child != nullptr) {
             child->InitFromType(refType, ds);
             type = child->type;
             name = child->name;
@@ -411,7 +406,7 @@ void Declarator::InitFromType(const Type *baseType, DeclSpecs *ds) {
         }
 
         const Type *arrayType = new ArrayType(baseType, arraySize);
-        if (child != NULL) {
+        if (child != nullptr) {
             child->InitFromType(arrayType, ds);
             type = child->type;
             name = child->name;
@@ -430,7 +425,7 @@ void Declarator::InitFromType(const Type *baseType, DeclSpecs *ds) {
         for (unsigned int i = 0; i < functionParams.size(); ++i) {
             Declaration *d = functionParams[i];
 
-            if (d == NULL) {
+            if (d == nullptr) {
                 AssertPos(pos, m->errorCount > 0);
                 continue;
             }
@@ -443,7 +438,7 @@ void Declarator::InitFromType(const Type *baseType, DeclSpecs *ds) {
 
             AssertPos(pos, d->declarators.size() == 1);
             Declarator *decl = d->declarators[0];
-            if (decl == NULL || decl->type == NULL) {
+            if (decl == nullptr || decl->type == nullptr) {
                 AssertPos(pos, m->errorCount > 0);
                 continue;
             }
@@ -454,7 +449,9 @@ void Declarator::InitFromType(const Type *baseType, DeclSpecs *ds) {
                 snprintf(buf, sizeof(buf), "__anon_parameter_%d", i);
                 decl->name = buf;
             }
-            decl->type = decl->type->ResolveUnboundVariability(Variability::Varying);
+            if (!decl->type->IsDependentType()) {
+                decl->type = decl->type->ResolveUnboundVariability(Variability::Varying);
+            }
 
             if (d->declSpecs->storageClass != SC_NONE)
                 Error(decl->pos,
@@ -464,11 +461,11 @@ void Declarator::InitFromType(const Type *baseType, DeclSpecs *ds) {
             if (decl->type->IsVoidType()) {
                 Error(decl->pos, "Parameter with type \"void\" illegal in function "
                                  "parameter list.");
-                decl->type = NULL;
+                decl->type = nullptr;
             }
 
             const ArrayType *at = CastType<ArrayType>(decl->type);
-            if (at != NULL) {
+            if (at != nullptr) {
                 // As in C, arrays are passed to functions as pointers to
                 // their element type.  We'll just immediately make this
                 // change now.  (One shortcoming of losing the fact that
@@ -478,7 +475,7 @@ void Declarator::InitFromType(const Type *baseType, DeclSpecs *ds) {
                 // in the function, but it's not clear that this is a
                 // significant problem.)
                 const Type *targetType = at->GetElementType();
-                if (targetType == NULL) {
+                if (targetType == nullptr) {
                     AssertPos(pos, m->errorCount > 0);
                     return;
                 }
@@ -488,7 +485,7 @@ void Declarator::InitFromType(const Type *baseType, DeclSpecs *ds) {
                 // Make sure there are no unsized arrays (other than the
                 // first dimension) in function parameter lists.
                 at = CastType<ArrayType>(targetType);
-                while (at != NULL) {
+                while (at != nullptr) {
                     if (at->GetElementCount() == 0)
                         Error(decl->pos, "Arrays with unsized dimensions in "
                                          "dimensions after the first one are illegal in "
@@ -501,17 +498,17 @@ void Declarator::InitFromType(const Type *baseType, DeclSpecs *ds) {
             argNames.push_back(decl->name);
             argPos.push_back(decl->pos);
 
-            Expr *init = NULL;
+            Expr *init = nullptr;
             // Try to find an initializer expression.
-            while (decl != NULL) {
-                if (decl->initExpr != NULL) {
+            while (decl != nullptr) {
+                if (decl->initExpr != nullptr) {
                     decl->initExpr = TypeCheck(decl->initExpr);
                     decl->initExpr = Optimize(decl->initExpr);
-                    if (decl->initExpr != NULL) {
+                    if (decl->initExpr != nullptr) {
                         init = llvm::dyn_cast<ConstExpr>(decl->initExpr);
-                        if (init == NULL)
+                        if (init == nullptr)
                             init = llvm::dyn_cast<NullPointerExpr>(decl->initExpr);
-                        if (init == NULL)
+                        if (init == nullptr)
                             Error(decl->initExpr->pos,
                                   "Default value for parameter "
                                   "\"%s\" must be a compile-time constant.",
@@ -525,22 +522,27 @@ void Declarator::InitFromType(const Type *baseType, DeclSpecs *ds) {
         }
 
         const Type *returnType = baseType;
-        if (returnType == NULL) {
+        if (returnType == nullptr) {
             Error(pos, "No return type provided in function declaration.");
             return;
         }
 
-        if (CastType<FunctionType>(returnType) != NULL) {
+        if (CastType<FunctionType>(returnType) != nullptr) {
             Error(pos, "Illegal to return function type from function.");
             return;
         }
 
-        returnType = returnType->ResolveUnboundVariability(Variability::Varying);
+        if (!returnType->IsDependentType()) {
+            returnType = returnType->ResolveUnboundVariability(Variability::Varying);
+        }
 
         bool isExternC = ds && (ds->storageClass == SC_EXTERN_C);
+        bool isExternSYCL = ds && (ds->storageClass == SC_EXTERN_SYCL);
         bool isExported = ds && ((ds->typeQualifiers & TYPEQUAL_EXPORT) != 0);
         bool isTask = ds && ((ds->typeQualifiers & TYPEQUAL_TASK) != 0);
         bool isUnmasked = ds && ((ds->typeQualifiers & TYPEQUAL_UNMASKED) != 0);
+        bool isVectorCall = ds && ((ds->typeQualifiers & TYPEQUAL_VECTORCALL) != 0);
+        bool isRegCall = ds && ((ds->typeQualifiers & TYPEQUAL_REGCALL) != 0);
 
         if (isExported && isTask) {
             Error(pos, "Function can't have both \"task\" and \"export\" "
@@ -557,20 +559,31 @@ void Declarator::InitFromType(const Type *baseType, DeclSpecs *ds) {
                        "qualifiers");
             return;
         }
+        if (isExternSYCL && isTask) {
+            Error(pos, "Function can't have both \"extern \"SYCL\"\" and \"task\" "
+                       "qualifiers");
+            return;
+        }
+        if (isExternSYCL && isExported) {
+            Error(pos, "Function can't have both \"extern \"SYCL\"\" and \"export\" "
+                       "qualifiers");
+            return;
+        }
         if (isUnmasked && isExported)
             Warning(pos, "\"unmasked\" qualifier is redundant for exported "
                          "functions.");
 
-        if (child == NULL) {
+        if (child == nullptr) {
             AssertPos(pos, m->errorCount > 0);
             return;
         }
 
-        const FunctionType *functionType = new FunctionType(returnType, args, argNames, argDefaults, argPos, isTask,
-                                                            isExported, isExternC, isUnmasked);
+        const FunctionType *functionType =
+            new FunctionType(returnType, args, argNames, argDefaults, argPos, isTask, isExported, isExternC,
+                             isExternSYCL, isUnmasked, isVectorCall, isRegCall);
 
         // handle any explicit __declspecs on the function
-        if (ds != NULL) {
+        if (ds != nullptr) {
             for (int i = 0; i < (int)ds->declSpecList.size(); ++i) {
                 std::string str = ds->declSpecList[i].first;
                 SourcePos ds_spec_pos = ds->declSpecList[i].second;
@@ -590,6 +603,8 @@ void Declarator::InitFromType(const Type *baseType, DeclSpecs *ds) {
         child->InitFromType(functionType, ds);
         type = child->type;
         name = child->name;
+    } else {
+        UNREACHABLE();
     }
 }
 
@@ -598,16 +613,16 @@ void Declarator::InitFromType(const Type *baseType, DeclSpecs *ds) {
 
 Declaration::Declaration(DeclSpecs *ds, std::vector<Declarator *> *dlist) {
     declSpecs = ds;
-    if (dlist != NULL)
+    if (dlist != nullptr)
         declarators = *dlist;
     for (unsigned int i = 0; i < declarators.size(); ++i)
-        if (declarators[i] != NULL)
+        if (declarators[i] != nullptr)
             declarators[i]->InitFromDeclSpecs(declSpecs);
 }
 
 Declaration::Declaration(DeclSpecs *ds, Declarator *d) {
     declSpecs = ds;
-    if (d != NULL) {
+    if (d != nullptr) {
         d->InitFromDeclSpecs(ds);
         declarators.push_back(d);
     }
@@ -619,7 +634,7 @@ std::vector<VariableDeclaration> Declaration::GetVariableDeclarations() const {
 
     for (unsigned int i = 0; i < declarators.size(); ++i) {
         Declarator *decl = declarators[i];
-        if (decl == NULL || decl->type == NULL) {
+        if (decl == nullptr || decl->type == nullptr) {
             // Ignore earlier errors
             Assert(m->errorCount > 0);
             continue;
@@ -627,11 +642,15 @@ std::vector<VariableDeclaration> Declaration::GetVariableDeclarations() const {
 
         if (decl->type->IsVoidType())
             Error(decl->pos, "\"void\" type variable illegal in declaration.");
-        else if (CastType<FunctionType>(decl->type) == NULL) {
-            decl->type = decl->type->ResolveUnboundVariability(Variability::Varying);
+        else if (CastType<FunctionType>(decl->type) == nullptr) {
+            if (!decl->type->IsDependentType()) {
+                decl->type = decl->type->ResolveUnboundVariability(Variability::Varying);
+            }
             Symbol *sym = new Symbol(decl->name, decl->pos, decl->type, decl->storageClass);
             m->symbolTable->AddVariable(sym);
             vars.push_back(VariableDeclaration(sym, decl->initExpr));
+        } else {
+            Error(decl->pos, "\"%s\" is illegal in declaration.", decl->name.c_str());
         }
     }
 
@@ -643,29 +662,43 @@ void Declaration::DeclareFunctions() {
 
     for (unsigned int i = 0; i < declarators.size(); ++i) {
         Declarator *decl = declarators[i];
-        if (decl == NULL || decl->type == NULL) {
+        if (decl == nullptr || decl->type == nullptr) {
             // Ignore earlier errors
             Assert(m->errorCount > 0);
             continue;
         }
 
         const FunctionType *ftype = CastType<FunctionType>(decl->type);
-        if (ftype == NULL)
+        if (ftype == nullptr)
             continue;
 
         bool isInline = (declSpecs->typeQualifiers & TYPEQUAL_INLINE);
         bool isNoInline = (declSpecs->typeQualifiers & TYPEQUAL_NOINLINE);
         bool isVectorCall = (declSpecs->typeQualifiers & TYPEQUAL_VECTORCALL);
-        m->AddFunctionDeclaration(decl->name, ftype, decl->storageClass, isInline, isNoInline, isVectorCall, decl->pos);
+        bool isRegCall = (declSpecs->typeQualifiers & TYPEQUAL_REGCALL);
+        m->AddFunctionDeclaration(decl->name, ftype, decl->storageClass, isInline, isNoInline, isVectorCall, isRegCall,
+                                  decl->pos);
     }
 }
 
-void Declaration::Print(int indent) const {
-    printf("%*cDeclaration: specs [", indent, ' ');
+void Declaration::Print() const {
+    Indent indent;
+    indent.pushSingle();
+    Print(indent);
+    fflush(stdout);
+}
+
+void Declaration::Print(Indent &indent) const {
+    indent.Print("Declaration: specs [");
     declSpecs->Print();
     printf("], declarators:\n");
-    for (unsigned int i = 0; i < declarators.size(); ++i)
-        declarators[i]->Print(indent + 4);
+
+    indent.pushList(declarators.size());
+    for (unsigned int i = 0; i < declarators.size(); ++i) {
+        declarators[i]->Print(indent);
+    }
+
+    indent.Done();
 }
 
 ///////////////////////////////////////////////////////////////////////////
@@ -677,7 +710,7 @@ void ispc::GetStructTypesNamesPositions(const std::vector<StructDeclaration *> &
     std::set<std::string> seenNames;
     for (unsigned int i = 0; i < sd.size(); ++i) {
         const Type *type = sd[i]->type;
-        if (type == NULL)
+        if (type == nullptr)
             continue;
 
         // FIXME: making this fake little DeclSpecs here is really
@@ -718,7 +751,7 @@ void ispc::GetStructTypesNamesPositions(const std::vector<StructDeclaration *> &
     for (int i = 0; i < (int)elementTypes->size() - 1; ++i) {
         const ArrayType *arrayType = CastType<ArrayType>((*elementTypes)[i]);
 
-        if (arrayType != NULL && arrayType->GetElementCount() == 0)
+        if (arrayType != nullptr && arrayType->GetElementCount() == 0)
             Error((*elementPositions)[i], "Unsized arrays aren't allowed except "
                                           "for the last member in a struct definition.");
     }
