@@ -1,34 +1,7 @@
 /*
-  Copyright (c) 2010-2021, Intel Corporation
-  All rights reserved.
+  Copyright (c) 2010-2024, Intel Corporation
 
-  Redistribution and use in source and binary forms, with or without
-  modification, are permitted provided that the following conditions are
-  met:
-
-    * Redistributions of source code must retain the above copyright
-      notice, this list of conditions and the following disclaimer.
-
-    * Redistributions in binary form must reproduce the above copyright
-      notice, this list of conditions and the following disclaimer in the
-      documentation and/or other materials provided with the distribution.
-
-    * Neither the name of Intel Corporation nor the names of its
-      contributors may be used to endorse or promote products derived from
-      this software without specific prior written permission.
-
-
-   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
-   IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
-   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
-   PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
-   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
-   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
-   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
-   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
-   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
-   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
-   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+  SPDX-License-Identifier: BSD-3-Clause
 */
 
 /** @file expr.h
@@ -38,6 +11,8 @@
 #pragma once
 
 #include "ast.h"
+#include "ctx.h"
+#include "func.h"
 #include "ispc.h"
 #include "type.h"
 
@@ -63,7 +38,7 @@ class Expr : public ASTNode {
         this function should emit IR that computes the expression's lvalue
         and returns the corresponding llvm::Value.  Expressions that can't
         provide an lvalue should leave this unimplemented; the default
-        implementation returns NULL.  */
+        implementation returns nullptr.  */
     virtual llvm::Value *GetLValue(FunctionEmitContext *ctx) const;
 
     /** Returns the Type of the expression. */
@@ -83,7 +58,7 @@ class Expr : public ASTNode {
         corresponding llvm::Constant value and a flag denoting if it's
         valid for multi-target compilation for use as an initializer of
         a global variable. Otherwise it should return the llvm::constant
-        value as NULL. */
+        value as nullptr. */
     virtual std::pair<llvm::Constant *, bool> GetStorageConstant(const Type *type) const;
 
     /** If this is a constant expression that can be converted to a
@@ -91,22 +66,21 @@ class Expr : public ASTNode {
         corresponding llvm::Constant value and a flag denoting if it's
         valid for multi-target compilation for use as an initializer of
         a global variable. Otherwise it should return the llvm::constant
-        value as NULL. */
+        value as nullptr. */
     virtual std::pair<llvm::Constant *, bool> GetConstant(const Type *type) const;
 
     /** This method should perform early optimizations of the expression
         (constant folding, etc.) and return a pointer to the resulting
-        expression.  If an error is encountered during optimization, NULL
+        expression.  If an error is encountered during optimization, nullptr
         should be returned. */
     virtual Expr *Optimize() = 0;
 
     /** This method should perform type checking of the expression and
         return a pointer to the resulting expression.  If an error is
-        encountered, NULL should be returned. */
+        encountered, nullptr should be returned. */
     virtual Expr *TypeCheck() = 0;
 
-    /** Prints the expression to standard output (used for debugging). */
-    virtual void Print() const = 0;
+    virtual Expr *Instantiate(TemplateInstantiation &templInst) const = 0;
 
     virtual bool HasAmbiguousVariability(std::vector<const Expr *> &warn) const;
 };
@@ -131,10 +105,11 @@ class UnaryExpr : public Expr {
 
     llvm::Value *GetValue(FunctionEmitContext *ctx) const;
     const Type *GetType() const;
-    void Print() const;
+    void Print(Indent &indent) const;
     Expr *Optimize();
     Expr *TypeCheck();
     int EstimateCost() const;
+    UnaryExpr *Instantiate(TemplateInstantiation &templInst) const;
 
     const Op op;
     Expr *expr;
@@ -176,11 +151,12 @@ class BinaryExpr : public Expr {
     llvm::Value *GetValue(FunctionEmitContext *ctx) const;
     const Type *GetType() const;
     const Type *GetLValueType() const;
-    void Print() const;
+    void Print(Indent &indent) const;
 
     Expr *Optimize();
     Expr *TypeCheck();
     int EstimateCost() const;
+    Expr *Instantiate(TemplateInstantiation &templInst) const;
     std::pair<llvm::Constant *, bool> GetStorageConstant(const Type *type) const;
     std::pair<llvm::Constant *, bool> GetConstant(const Type *type) const;
     bool HasAmbiguousVariability(std::vector<const Expr *> &warn) const;
@@ -213,11 +189,12 @@ class AssignExpr : public Expr {
 
     llvm::Value *GetValue(FunctionEmitContext *ctx) const;
     const Type *GetType() const;
-    void Print() const;
+    void Print(Indent &indent) const;
 
     Expr *Optimize();
     Expr *TypeCheck();
     int EstimateCost() const;
+    AssignExpr *Instantiate(TemplateInstantiation &templInst) const;
 
     const Op op;
     Expr *lvalue, *rvalue;
@@ -236,11 +213,13 @@ class SelectExpr : public Expr {
 
     llvm::Value *GetValue(FunctionEmitContext *ctx) const;
     const Type *GetType() const;
-    void Print() const;
+    const Type *GetLValueType() const;
+    void Print(Indent &indent) const;
 
     Expr *Optimize();
     Expr *TypeCheck();
     int EstimateCost() const;
+    SelectExpr *Instantiate(TemplateInstantiation &templInst) const;
     bool HasAmbiguousVariability(std::vector<const Expr *> &warn) const;
 
     Expr *test, *expr1, *expr2;
@@ -262,22 +241,45 @@ class ExprList : public Expr {
 
     llvm::Value *GetValue(FunctionEmitContext *ctx) const;
     const Type *GetType() const;
-    void Print() const;
+    void Print(Indent &indent) const;
     std::pair<llvm::Constant *, bool> GetStorageConstant(const Type *type) const;
     std::pair<llvm::Constant *, bool> GetConstant(const Type *type) const;
     ExprList *Optimize();
     ExprList *TypeCheck();
     int EstimateCost() const;
+    ExprList *Instantiate(TemplateInstantiation &templInst) const;
     bool HasAmbiguousVariability(std::vector<const Expr *> &warn) const;
 
     std::vector<Expr *> exprs;
+
+    // Utility structs to support initializers lists for vectors
+    struct ExprPosMapping {
+        Expr *expr;
+        int pos;
+        ExprPosMapping(Expr *e, int p) : expr(e), pos(p) {}
+    };
+
+    struct ExprPosMappingVectorForVector {
+        int vec_mem_pos_from;
+        int vec_mem_pos_to;
+        Expr *expr;
+        ExprPosMappingVectorForVector(int from, int to, Expr *e)
+            : vec_mem_pos_from(from), vec_mem_pos_to(to), expr(e) {}
+        ExprPosMappingVectorForVector(int from, ExprPosMapping map)
+            : vec_mem_pos_from(from), vec_mem_pos_to(map.pos), expr(map.expr) {}
+    };
+
+    // Returns true if each expression in expression list has AtomicType.
+    // It also constructs a map of initializers for each atomic basetype.
+    bool HasAtomicInitializerList(std::map<AtomicType::BasicType, std::vector<ExprPosMapping>> &map);
 };
 
 /** @brief Expression representing a function call.
  */
 class FunctionCallExpr : public Expr {
   public:
-    FunctionCallExpr(Expr *func, ExprList *args, SourcePos p, bool isLaunch = false, Expr *launchCountExpr[3] = NULL);
+    FunctionCallExpr(Expr *func, ExprList *args, SourcePos p, bool isLaunch = false, Expr *launchCountExpr[3] = nullptr,
+                     bool isInvoke = false);
 
     static inline bool classof(FunctionCallExpr const *) { return true; }
     static inline bool classof(ASTNode const *N) { return N->getValueID() == FunctionCallExprID; }
@@ -286,15 +288,17 @@ class FunctionCallExpr : public Expr {
     llvm::Value *GetLValue(FunctionEmitContext *ctx) const;
     const Type *GetType() const;
     const Type *GetLValueType() const;
-    void Print() const;
+    void Print(Indent &indent) const;
 
     Expr *Optimize();
     Expr *TypeCheck();
     int EstimateCost() const;
+    FunctionCallExpr *Instantiate(TemplateInstantiation &templInst) const;
 
     Expr *func;
     ExprList *args;
     bool isLaunch;
+    bool isInvoke;
     Expr *launchCountExpr[3];
 };
 
@@ -315,11 +319,12 @@ class IndexExpr : public Expr {
     const Type *GetType() const;
     const Type *GetLValueType() const;
     Symbol *GetBaseSymbol() const;
-    void Print() const;
+    void Print(Indent &indent) const;
 
     Expr *Optimize();
     Expr *TypeCheck();
     int EstimateCost() const;
+    IndexExpr *Instantiate(TemplateInstantiation &templInst) const;
 
     Expr *baseExpr, *index;
 
@@ -339,17 +344,19 @@ class MemberExpr : public Expr {
 
     static inline bool classof(MemberExpr const *) { return true; }
     static inline bool classof(ASTNode const *N) {
-        return ((N->getValueID() == StructMemberExprID) || (N->getValueID() == VectorMemberExprID));
+        return ((N->getValueID() == StructMemberExprID) || (N->getValueID() == VectorMemberExprID) ||
+                (N->getValueID() == DependentMemberExprID));
     }
 
     llvm::Value *GetValue(FunctionEmitContext *ctx) const;
     llvm::Value *GetLValue(FunctionEmitContext *ctx) const;
     const Type *GetType() const;
     Symbol *GetBaseSymbol() const;
-    void Print() const;
+    void Print(Indent &indent) const;
     Expr *Optimize();
     Expr *TypeCheck();
     int EstimateCost() const;
+    MemberExpr *Instantiate(TemplateInstantiation &templInst) const;
 
     virtual int getElementNumber() const = 0;
     virtual const Type *getElementType() const = 0;
@@ -371,6 +378,53 @@ class MemberExpr : public Expr {
     mutable const Type *type, *lvalueType;
 };
 
+class StructMemberExpr : public MemberExpr {
+  public:
+    StructMemberExpr(Expr *e, const char *id, SourcePos p, SourcePos idpos, bool derefLValue);
+
+    static inline bool classof(StructMemberExpr const *) { return true; }
+    static inline bool classof(ASTNode const *N) { return N->getValueID() == StructMemberExprID; }
+
+    const Type *GetType() const;
+    const Type *GetLValueType() const;
+    int getElementNumber() const;
+    const Type *getElementType() const;
+
+  private:
+    const StructType *getStructType() const;
+};
+
+class VectorMemberExpr : public MemberExpr {
+  public:
+    VectorMemberExpr(Expr *e, const char *id, SourcePos p, SourcePos idpos, bool derefLValue);
+
+    static inline bool classof(VectorMemberExpr const *) { return true; }
+    static inline bool classof(ASTNode const *N) { return N->getValueID() == VectorMemberExprID; }
+
+    llvm::Value *GetValue(FunctionEmitContext *ctx) const;
+    llvm::Value *GetLValue(FunctionEmitContext *ctx) const;
+    const Type *GetType() const;
+    const Type *GetLValueType() const;
+
+    int getElementNumber() const;
+    const Type *getElementType() const;
+
+  private:
+    const VectorType *exprVectorType;
+    const VectorType *memberType;
+};
+
+class DependentMemberExpr : public MemberExpr {
+  public:
+    DependentMemberExpr(Expr *e, const char *id, SourcePos p, SourcePos idpos, bool derefLValue);
+
+    static inline bool classof(DependentMemberExpr const *) { return true; }
+    static inline bool classof(ASTNode const *N) { return N->getValueID() == DependentMemberExprID; }
+
+    int getElementNumber() const;
+    const Type *getElementType() const;
+};
+
 /** @brief Expression representing a compile-time constant value.
 
     This class can currently represent compile-time constants of anything
@@ -407,15 +461,10 @@ class ConstExpr : public Expr {
     /** Create a ConstExpr from a varying uint32 value */
     ConstExpr(const Type *t, uint32_t *u, SourcePos p);
 
-    /** Create a ConstExpr from a uniform float value */
-    ConstExpr(const Type *t, float f, SourcePos p);
-    /** Create a ConstExpr from a varying float value */
-    ConstExpr(const Type *t, float *f, SourcePos p);
-
-    /** Create a ConstExpr from a uniform double value */
-    ConstExpr(const Type *t, double d, SourcePos p);
-    /** Create a ConstExpr from a varying double value */
-    ConstExpr(const Type *t, double *d, SourcePos p);
+    /** Create a ConstExpr from a llvm::APFloat value for uniform floating point types */
+    ConstExpr(const Type *t, llvm::APFloat f, SourcePos p);
+    /** Create a ConstExpr from a llvm::APFloat value for varying floating point types */
+    ConstExpr(const Type *t, std::vector<llvm::APFloat> const &f, SourcePos p);
 
     /** Create a ConstExpr from a uniform int64 value */
     ConstExpr(const Type *t, int64_t i, SourcePos p);
@@ -431,26 +480,23 @@ class ConstExpr : public Expr {
     /** Create a ConstExpr from a varying bool value */
     ConstExpr(const Type *t, bool *b, SourcePos p);
 
-    /** Create a ConstExpr of the same type as the given old ConstExpr,
-        with values given by the "vales" parameter. */
-    ConstExpr(ConstExpr *old, double *values);
-
     /** Create ConstExpr with the same type and values as the given one,
         but at the given position. */
-    ConstExpr(ConstExpr *old, SourcePos pos);
+    ConstExpr(const ConstExpr *old, SourcePos pos);
 
     static inline bool classof(ConstExpr const *) { return true; }
     static inline bool classof(ASTNode const *N) { return N->getValueID() == ConstExprID; }
 
     llvm::Value *GetValue(FunctionEmitContext *ctx) const;
     const Type *GetType() const;
-    void Print() const;
+    void Print(Indent &indent) const;
     std::pair<llvm::Constant *, bool> GetStorageConstant(const Type *type) const;
     std::pair<llvm::Constant *, bool> GetConstant(const Type *constType) const;
 
     Expr *TypeCheck();
     Expr *Optimize();
     int EstimateCost() const;
+    ConstExpr *Instantiate(TemplateInstantiation &templInst) const;
 
     /** Return the ConstExpr's values as the given pointer type, doing type
         conversion from the actual type if needed.  If forceVarying is
@@ -462,18 +508,24 @@ class ConstExpr : public Expr {
     int GetValues(uint8_t *, bool forceVarying = false) const;
     int GetValues(int16_t *, bool forceVarying = false) const;
     int GetValues(uint16_t *, bool forceVarying = false) const;
+    int GetValues(std::vector<llvm::APFloat> &, llvm::Type *, bool forceVarying = false) const;
     int GetValues(int32_t *, bool forceVarying = false) const;
     int GetValues(uint32_t *, bool forceVarying = false) const;
-    int GetValues(float *, bool forceVarying = false) const;
     int GetValues(int64_t *, bool forceVarying = false) const;
     int GetValues(uint64_t *, bool forceVarying = false) const;
-    int GetValues(double *, bool forceVarying = false) const;
+    int GetValues(std::vector<llvm::APFloat> &) const;
+
+    /** Return the ConstExpr's values as a string. */
+    std::string GetValuesAsStr(const std::string &separator) const;
 
     /** Return the number of values in the ConstExpr; should be either 1,
         if it has uniform type, or the target's vector width if it's
         varying. */
     int Count() const;
 
+    /** Return true if the type and values of two ConstExpr are the same. */
+    bool IsEqual(const ConstExpr *ce) const;
+
   private:
     AtomicType::BasicType getBasicType() const;
 
@@ -486,11 +538,10 @@ class ConstExpr : public Expr {
         int32_t int32Val[ISPC_MAX_NVEC];
         uint32_t uint32Val[ISPC_MAX_NVEC];
         bool boolVal[ISPC_MAX_NVEC];
-        float floatVal[ISPC_MAX_NVEC];
-        double doubleVal[ISPC_MAX_NVEC];
         int64_t int64Val[ISPC_MAX_NVEC];
         uint64_t uint64Val[ISPC_MAX_NVEC];
     };
+    std::vector<llvm::APFloat> fpVal;
 };
 
 /** @brief Expression representing a type cast of the given expression to a
@@ -506,10 +557,11 @@ class TypeCastExpr : public Expr {
     llvm::Value *GetLValue(FunctionEmitContext *ctx) const;
     const Type *GetType() const;
     const Type *GetLValueType() const;
-    void Print() const;
+    void Print(Indent &indent) const;
     Expr *TypeCheck();
     Expr *Optimize();
     int EstimateCost() const;
+    TypeCastExpr *Instantiate(TemplateInstantiation &templInst) const;
     Symbol *GetBaseSymbol() const;
     std::pair<llvm::Constant *, bool> GetConstant(const Type *type) const;
     bool HasAmbiguousVariability(std::vector<const Expr *> &warn) const;
@@ -532,10 +584,11 @@ class ReferenceExpr : public Expr {
     const Type *GetType() const;
     const Type *GetLValueType() const;
     Symbol *GetBaseSymbol() const;
-    void Print() const;
+    void Print(Indent &indent) const;
     Expr *TypeCheck();
     Expr *Optimize();
     int EstimateCost() const;
+    ReferenceExpr *Instantiate(TemplateInstantiation &templInst) const;
 
     Expr *expr;
 };
@@ -571,9 +624,10 @@ class PtrDerefExpr : public DerefExpr {
     static inline bool classof(ASTNode const *N) { return N->getValueID() == PtrDerefExprID; }
 
     const Type *GetType() const;
-    void Print() const;
+    void Print(Indent &indent) const;
     Expr *TypeCheck();
     int EstimateCost() const;
+    PtrDerefExpr *Instantiate(TemplateInstantiation &templInst) const;
 };
 
 /** @brief Expression that represents dereferencing a reference to get its
@@ -586,9 +640,10 @@ class RefDerefExpr : public DerefExpr {
     static inline bool classof(ASTNode const *N) { return N->getValueID() == RefDerefExprID; }
 
     const Type *GetType() const;
-    void Print() const;
+    void Print(Indent &indent) const;
     Expr *TypeCheck();
     int EstimateCost() const;
+    RefDerefExpr *Instantiate(TemplateInstantiation &templInst) const;
 };
 
 /** Expression that represents taking the address of an expression. */
@@ -603,10 +658,11 @@ class AddressOfExpr : public Expr {
     const Type *GetType() const;
     const Type *GetLValueType() const;
     Symbol *GetBaseSymbol() const;
-    void Print() const;
+    void Print(Indent &indent) const;
     Expr *TypeCheck();
     Expr *Optimize();
     int EstimateCost() const;
+    AddressOfExpr *Instantiate(TemplateInstantiation &templInst) const;
     std::pair<llvm::Constant *, bool> GetConstant(const Type *type) const;
 
     Expr *expr;
@@ -624,14 +680,15 @@ class SizeOfExpr : public Expr {
 
     llvm::Value *GetValue(FunctionEmitContext *ctx) const;
     const Type *GetType() const;
-    void Print() const;
+    void Print(Indent &indent) const;
     Expr *TypeCheck();
     Expr *Optimize();
     int EstimateCost() const;
+    SizeOfExpr *Instantiate(TemplateInstantiation &templInst) const;
     std::pair<llvm::Constant *, bool> GetConstant(const Type *type) const;
 
-    /* One of expr or type should be non-NULL (but not both of them).  The
-       SizeOfExpr returns the size of whichever one of them isn't NULL. */
+    /* One of expr or type should be non-nullptr (but not both of them).  The
+       SizeOfExpr returns the size of whichever one of them isn't nullptr. */
     Expr *expr;
     const Type *type;
 };
@@ -647,10 +704,11 @@ class AllocaExpr : public Expr {
 
     llvm::Value *GetValue(FunctionEmitContext *ctx) const;
     const Type *GetType() const;
-    void Print() const;
+    void Print(Indent &indent) const;
     Expr *TypeCheck();
     Expr *Optimize();
     int EstimateCost() const;
+    AllocaExpr *Instantiate(TemplateInstantiation &templInst) const;
 
     // The expr should have size_t type and should evaluate to size
     // of stack memory to be allocated.
@@ -672,8 +730,9 @@ class SymbolExpr : public Expr {
     Symbol *GetBaseSymbol() const;
     Expr *TypeCheck();
     Expr *Optimize();
-    void Print() const;
+    void Print(Indent &indent) const;
     int EstimateCost() const;
+    SymbolExpr *Instantiate(TemplateInstantiation &templInst) const;
 
   private:
     Symbol *symbol;
@@ -685,6 +744,8 @@ class SymbolExpr : public Expr {
 class FunctionSymbolExpr : public Expr {
   public:
     FunctionSymbolExpr(const char *name, const std::vector<Symbol *> &candFuncs, SourcePos pos);
+    FunctionSymbolExpr(const char *name, const std::vector<TemplateSymbol *> &candFuncs, const TemplateArgs &templArgs,
+                       SourcePos pos);
 
     static inline bool classof(FunctionSymbolExpr const *) { return true; }
     static inline bool classof(ASTNode const *N) { return N->getValueID() == FunctionSymbolExprID; }
@@ -694,29 +755,30 @@ class FunctionSymbolExpr : public Expr {
     Symbol *GetBaseSymbol() const;
     Expr *TypeCheck();
     Expr *Optimize();
-    void Print() const;
+    void Print(Indent &indent) const;
     int EstimateCost() const;
+    FunctionSymbolExpr *Instantiate(TemplateInstantiation &templInst) const;
     std::pair<llvm::Constant *, bool> GetConstant(const Type *type) const;
 
     /** Given the types of the function arguments, in the presence of
         function overloading, this method resolves which actual function
         the arguments match best.  If the argCouldBeNULL parameter is
-        non-NULL, each element indicates whether the corresponding argument
-        is the number zero, indicating that it could be a NULL pointer, and
-        if argIsConstant is non-NULL, each element indicates whether the
+        non-nullptr, each element indicates whether the corresponding argument
+        is the number zero, indicating that it could be a nullptr pointer, and
+        if argIsConstant is non-nullptr, each element indicates whether the
         corresponding argument is a compile-time constant value.  Both of
-        these parameters may be NULL (for cases where overload resolution
+        these parameters may be nullptr (for cases where overload resolution
         is being done just given type information without the parameter
         argument expressions being available.  This function returns true
         on success.
      */
     bool ResolveOverloads(SourcePos argPos, const std::vector<const Type *> &argTypes,
-                          const std::vector<bool> *argCouldBeNULL = NULL,
-                          const std::vector<bool> *argIsConstant = NULL);
-    Symbol *GetMatchingFunction();
+                          const std::vector<bool> *argCouldBeNULL = nullptr,
+                          const std::vector<bool> *argIsConstant = nullptr);
 
   private:
     std::vector<Symbol *> getCandidateFunctions(int argCount) const;
+    std::vector<Symbol *> getCandidateTemplateFunctions(const std::vector<const Type *> &argTypes) const;
     static int computeOverloadCost(const FunctionType *ftype, const std::vector<const Type *> &argTypes,
                                    const std::vector<bool> *argCouldBeNULL, const std::vector<bool> *argIsConstant,
                                    int *cost);
@@ -728,11 +790,14 @@ class FunctionSymbolExpr : public Expr {
         there may be more then one, in which case we need to resolve which
         overload is the best match. */
     std::vector<Symbol *> candidateFunctions;
+    std::vector<TemplateSymbol *> candidateTemplateFunctions;
+    TemplateArgs templateArgs;
 
     /** The actual matching function found after overload resolution. */
     Symbol *matchingFunc;
 
     bool triedToResolve;
+    bool unresolvedButDependent;
 };
 
 /** @brief A sync statement in the program (waits for all launched tasks before
@@ -748,11 +813,12 @@ class SyncExpr : public Expr {
     const Type *GetType() const;
     Expr *TypeCheck();
     Expr *Optimize();
-    void Print() const;
+    void Print(Indent &indent) const;
     int EstimateCost() const;
+    SyncExpr *Instantiate(TemplateInstantiation &templInst) const;
 };
 
-/** @brief An expression that represents a NULL pointer. */
+/** @brief An expression that represents a nullptr pointer. */
 class NullPointerExpr : public Expr {
   public:
     NullPointerExpr(SourcePos p) : Expr(p, NullPointerExprID) {}
@@ -765,8 +831,9 @@ class NullPointerExpr : public Expr {
     Expr *TypeCheck();
     Expr *Optimize();
     std::pair<llvm::Constant *, bool> GetConstant(const Type *type) const;
-    void Print() const;
+    void Print(Indent &indent) const;
     int EstimateCost() const;
+    NullPointerExpr *Instantiate(TemplateInstantiation &templInst) const;
 };
 
 /** An expression representing a "new" expression, used for dynamically
@@ -783,13 +850,14 @@ class NewExpr : public Expr {
     const Type *GetType() const;
     Expr *TypeCheck();
     Expr *Optimize();
-    void Print() const;
+    void Print(Indent &indent) const;
     int EstimateCost() const;
+    NewExpr *Instantiate(TemplateInstantiation &templInst) const;
 
     /** Type of object to allocate storage for. */
     const Type *allocType;
     /** Expression giving the number of elements to allocate, when the
-        "new Foo[expr]" form is used.  This may be NULL, in which case a
+        "new Foo[expr]" form is used.  This may be nullptr, in which case a
         single element of the given type will be allocated. */
     Expr *countExpr;
     /** Optional initializer expression used to initialize the allocated
@@ -800,6 +868,9 @@ class NewExpr : public Expr {
         instance, or whether a single allocation is performed for the
         entire gang of program instances.) */
     bool isVarying;
+
+  private:
+    NewExpr(const Type *type, Expr *count, Expr *init, bool isV, SourcePos p);
 };
 
 /** This function indicates whether it's legal to convert from fromType to
@@ -807,12 +878,12 @@ class NewExpr : public Expr {
     are provided, then an error message is issued if the type conversion
     isn't possible.
  */
-bool CanConvertTypes(const Type *fromType, const Type *toType, const char *errorMsgBase = NULL,
+bool CanConvertTypes(const Type *fromType, const Type *toType, const char *errorMsgBase = nullptr,
                      SourcePos pos = SourcePos());
 
 /** This function attempts to convert the given expression to the given
     type, returning a pointer to a new expression that is the result.  If
-    the required type conversion is illegal, it returns NULL and prints an
+    the required type conversion is illegal, it returns nullptr and prints an
     error message using the provided string to indicate the context for
     which type conversion was being applied (e.g. "function call
     parameter").
@@ -831,7 +902,7 @@ Expr *MakeBinaryExpr(BinaryExpr::Op o, Expr *a, Expr *b, SourcePos p);
     @param ctx       FunctionEmitContext to use for generating instructions
     @param pos       Source file position of the variable being initialized
 */
-void InitSymbol(llvm::Value *lvalue, const Type *symType, Expr *initExpr, FunctionEmitContext *ctx, SourcePos pos);
+void InitSymbol(AddressInfo *lvalue, const Type *symType, Expr *initExpr, FunctionEmitContext *ctx, SourcePos pos);
 
 bool PossiblyResolveFunctionOverloads(Expr *expr, const Type *type);
 } // namespace ispc
