@@ -1,34 +1,7 @@
 /*
-  Copyright (c) 2010-2021, Intel Corporation
-  All rights reserved.
+  Copyright (c) 2010-2024, Intel Corporation
 
-  Redistribution and use in source and binary forms, with or without
-  modification, are permitted provided that the following conditions are
-  met:
-
-    * Redistributions of source code must retain the above copyright
-      notice, this list of conditions and the following disclaimer.
-
-    * Redistributions in binary form must reproduce the above copyright
-      notice, this list of conditions and the following disclaimer in the
-      documentation and/or other materials provided with the distribution.
-
-    * Neither the name of Intel Corporation nor the names of its
-      contributors may be used to endorse or promote products derived from
-      this software without specific prior written permission.
-
-
-   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
-   IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
-   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
-   PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
-   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
-   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
-   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
-   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
-   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
-   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
-   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+  SPDX-License-Identifier: BSD-3-Clause
 */
 
 /** @file llvmutil.h
@@ -37,28 +10,39 @@
 
 #pragma once
 
+#include "ispc.h"
 #include "ispc_version.h"
 
 #include <llvm/IR/Constants.h>
 #include <llvm/IR/DerivedTypes.h>
+#include <llvm/IR/Instructions.h>
 #include <llvm/IR/LLVMContext.h>
 #include <llvm/IR/Type.h>
+#if ISPC_LLVM_VERSION >= ISPC_LLVM_16_0
+#include <llvm/Support/ModRef.h>
+#endif
 
-#define PTYPE(p) (llvm::cast<llvm::PointerType>((p)->getType()->getScalarType())->getElementType())
+// In the transition to Opaque Pointers getElementType() was deprecated, getPointerElementType() will live a little
+// longer. But we need another solution eventually. Issue #2245 was filed to track this.
+#if ISPC_LLVM_VERSION >= ISPC_LLVM_14_0
+#define PTR_ELT_TYPE getPointerElementType
+#else
+#define PTR_ELT_TYPE getElementType
+#endif
+
+#define PTYPE(p) (llvm::cast<llvm::PointerType>((p)->getType()->getScalarType())->PTR_ELT_TYPE())
 
 namespace llvm {
 class PHINode;
 class InsertElementInst;
 } // namespace llvm
 
-#if ISPC_LLVM_VERSION >= ISPC_LLVM_11_0
 #define LLVMVECTOR llvm::FixedVectorType
-#else
-#define LLVMVECTOR llvm::VectorType
-#endif
 
 namespace ispc {
 
+enum class MaskStatus { all_on, all_off, mixed, unknown };
+
 /** This structure holds pointers to a variety of LLVM types; code
     elsewhere can use them from here, ratherthan needing to make more
     verbose LLVM API calls.
@@ -74,6 +58,7 @@ struct LLVMTypes {
     static llvm::Type *Int16Type;
     static llvm::Type *Int32Type;
     static llvm::Type *Int64Type;
+    static llvm::Type *Float16Type;
     static llvm::Type *FloatType;
     static llvm::Type *DoubleType;
 
@@ -81,6 +66,7 @@ struct LLVMTypes {
     static llvm::Type *Int16PointerType;
     static llvm::Type *Int32PointerType;
     static llvm::Type *Int64PointerType;
+    static llvm::Type *Float16PointerType;
     static llvm::Type *FloatPointerType;
     static llvm::Type *DoublePointerType;
 
@@ -93,6 +79,7 @@ struct LLVMTypes {
     static llvm::VectorType *Int16VectorType;
     static llvm::VectorType *Int32VectorType;
     static llvm::VectorType *Int64VectorType;
+    static llvm::VectorType *Float16VectorType;
     static llvm::VectorType *FloatVectorType;
     static llvm::VectorType *DoubleVectorType;
 
@@ -100,6 +87,7 @@ struct LLVMTypes {
     static llvm::Type *Int16VectorPointerType;
     static llvm::Type *Int32VectorPointerType;
     static llvm::Type *Int64VectorPointerType;
+    static llvm::Type *Float16VectorPointerType;
     static llvm::Type *FloatVectorPointerType;
     static llvm::Type *DoubleVectorPointerType;
 
@@ -134,10 +122,12 @@ extern llvm::ConstantInt *LLVMUInt32(uint32_t i);
 extern llvm::ConstantInt *LLVMInt64(int64_t i);
 /** Returns an LLVM i64 constant of the given value */
 extern llvm::ConstantInt *LLVMUInt64(uint64_t i);
+/** Returns an LLVM half constant of the given value */
+extern llvm::Constant *LLVMFloat16(llvm::APFloat f);
 /** Returns an LLVM float constant of the given value */
-extern llvm::Constant *LLVMFloat(float f);
+extern llvm::Constant *LLVMFloat(llvm::APFloat f);
 /** Returns an LLVM double constant of the given value */
-extern llvm::Constant *LLVMDouble(double f);
+extern llvm::Constant *LLVMDouble(llvm::APFloat f);
 
 /** Returns an LLVM boolean vector constant of the given value smeared
     across all elements */
@@ -175,12 +165,15 @@ extern llvm::Constant *LLVMInt64Vector(int64_t i);
     across all elements */
 extern llvm::Constant *LLVMUInt64Vector(uint64_t i);
 
+/** Returns an LLVM half vector constant of the given value smeared
+    across all elements */
+extern llvm::Constant *LLVMFloat16Vector(llvm::APFloat f);
 /** Returns an LLVM float vector constant of the given value smeared
     across all elements */
-extern llvm::Constant *LLVMFloatVector(float f);
+extern llvm::Constant *LLVMFloatVector(llvm::APFloat f);
 /** Returns an LLVM double vector constant of the given value smeared
     across all elements */
-extern llvm::Constant *LLVMDoubleVector(double f);
+extern llvm::Constant *LLVMDoubleVector(llvm::APFloat f);
 
 /** Returns a constant integer or vector (according to the given type) of
     the given signed integer value. */
@@ -190,6 +183,9 @@ extern llvm::Constant *LLVMIntAsType(int64_t, llvm::Type *t);
     the given unsigned integer value. */
 extern llvm::Constant *LLVMUIntAsType(uint64_t, llvm::Type *t);
 
+/** Returns a zero constant half/float/double or vector (according to the given type). */
+extern llvm::Constant *LLVMFPZeroAsType(llvm::Type *type);
+
 /** Returns an LLVM boolean vector based on the given array of values.
     The array should have g->target.vectorWidth elements. */
 extern llvm::Constant *LLVMBoolVector(const bool *v);
@@ -227,12 +223,15 @@ extern llvm::Constant *LLVMInt64Vector(const int64_t *i);
     The array should have g->target.vectorWidth elements. */
 extern llvm::Constant *LLVMUInt64Vector(const uint64_t *i);
 
+/** Returns an LLVM half vector based on the given array of values.
+    The array should have g->target.vectorWidth elements. */
+extern llvm::Constant *LLVMFloat16Vector(const std::vector<llvm::APFloat> &f);
 /** Returns an LLVM float vector based on the given array of values.
     The array should have g->target.vectorWidth elements. */
-extern llvm::Constant *LLVMFloatVector(const float *f);
+extern llvm::Constant *LLVMFloatVector(const std::vector<llvm::APFloat> &f);
 /** Returns an LLVM double vector based on the given array of values.
     The array should have g->target.vectorWidth elements. */
-extern llvm::Constant *LLVMDoubleVector(const double *f);
+extern llvm::Constant *LLVMDoubleVector(const std::vector<llvm::APFloat> &f);
 
 /** LLVM constant value representing an 'all on' SIMD lane mask */
 extern llvm::Constant *LLVMMaskAllOn;
@@ -242,7 +241,7 @@ extern llvm::Constant *LLVMMaskAllOff;
 /** Tests to see if all of the elements of the vector in the 'v' parameter
     are equal.  Like lValuesAreEqual(), this is a conservative test and may
     return false for arrays where the values are actually all equal.  */
-extern bool LLVMVectorValuesAllEqual(llvm::Value *v, llvm::Value **splat = NULL);
+extern bool LLVMVectorValuesAllEqual(llvm::Value *v, llvm::Value **splat = nullptr);
 
 /** Tests to see if OR is actually an ADD.  */
 extern bool IsOrEquivalentToAdd(llvm::Value *op);
@@ -291,11 +290,11 @@ extern bool LLVMExtractVectorInts(llvm::Value *v, int64_t ret[], int *nElts);
                   <i64 4, i64 undef, i64 undef, i64 undef, i64 undef, i64 undef, i64 undef, i64 undef>
          %gep_offset = shufflevector <8 x i64> %0, <8 x i64> undef, <8 x i32> zeroinitializer
     Function returns:
-    Compare all elements and return one of them if all are equal, otherwise NULL.
+    Compare all elements and return one of them if all are equal, otherwise nullptr.
     If searchFirstUndef argument is true, look for the vector with the first not-undef element, like:
          <i64 4, i64 undef, i64 undef, i64 undef, i64 undef, i64 undef, i64 undef, i64 undef>
     If compare argument is false, don't do compare and return first element instead.
-    If undef argument is true, ignore undef elements (but all undef yields NULL anyway).
+    If undef argument is true, ignore undef elements (but all undef yields nullptr anyway).
 
  */
 extern llvm::Value *LLVMFlattenInsertChain(llvm::Value *inst, int vectorWidth, bool compare = true, bool undef = true,
@@ -304,9 +303,7 @@ extern llvm::Value *LLVMFlattenInsertChain(llvm::Value *inst, int vectorWidth, b
 /** This is a utility routine for debugging that dumps out the given LLVM
     value as well as (recursively) all of the other values that it depends
     on. */
-#ifndef ISPC_NO_DUMPS
 extern void LLVMDumpValue(llvm::Value *v);
-#endif
 
 /** Given a vector-typed value, this function returns the value of its
     first element.  Rather than just doing the straightforward thing of
@@ -330,9 +327,105 @@ extern llvm::Value *LLVMConcatVectors(llvm::Value *v1, llvm::Value *v2, llvm::In
 extern llvm::Value *LLVMShuffleVectors(llvm::Value *v1, llvm::Value *v2, int32_t shuf[], int shufSize,
                                        llvm::Instruction *insertBefore);
 
-#ifdef ISPC_GENX_ENABLED
-enum AddressSpace { Local, Global, External };
+/** This utility routine copies the metadata (if any) attached to the
+    'from' instruction in the IR to the 'to' instruction.
 
+    For flexibility, this function takes an llvm::Value rather than an
+    llvm::Instruction for the 'to' parameter; at some places in the code
+    below, we sometimes use a llvm::Value to start out storing a value and
+    then later store instructions.  If a llvm::Value is passed to this, the
+    routine just returns without doing anything; if it is in fact an
+    LLVM::Instruction, then the metadata can be copied to it. */
+extern void LLVMCopyMetadata(llvm::Value *vto, const llvm::Instruction *from);
+
+/** We have a protocol with the front-end LLVM IR code generation process
+    that allows us to encode the source file position that corresponds with
+    instructions.  (For example, this allows us to issue performance
+    warnings related to things like scatter and gather after optimization
+    has been performed, so that we aren't warning about scatters and
+    gathers that have been improved to stores and loads by optimization
+    passes.)  Note that this is slightly redundant with the source file
+    position encoding generated for debugging symbols, though we don't
+    always generate debugging information but we do always generate this
+    position data.
+
+    This function finds the SourcePos that the metadata in the instruction
+    (if present) corresponds to.  See the implementation of
+    FunctionEmitContext::addGSMetadata(), which encodes the source position during
+    code generation.
+
+    @param inst   Instruction to try to find the source position of
+    @param pos    Output variable in which to store the position
+    @returns      True if source file position metadata was present and *pos
+                  has been set.  False otherwise.*/
+extern bool LLVMGetSourcePosFromMetadata(const llvm::Instruction *inst, SourcePos *pos);
+
+/** Given an llvm::Value, return true if we can determine that it's an
+    undefined value.  This only makes a weak attempt at chasing this down,
+    only detecting flat-out undef values, and bitcasts of undef values.
+
+    @todo Is it worth working harder to find more of these?  It starts to
+    get tricky, since having an undef operand doesn't necessarily mean that
+    the result will be undefined.  (And for that matter, is there an LLVM
+    call that will do this for us?)
+ */
+extern bool LLVMIsValueUndef(llvm::Value *value);
+
+/** Below are helper functions to construct LLVM instructions. */
+extern llvm::CallInst *LLVMCallInst(llvm::Function *func, llvm::Value *arg0, llvm::Value *arg1, const llvm::Twine &name,
+                                    llvm::Instruction *insertBefore = nullptr);
+
+extern llvm::CallInst *LLVMCallInst(llvm::Function *func, llvm::Value *arg0, llvm::Value *arg1, llvm::Value *arg2,
+                                    const llvm::Twine &name, llvm::Instruction *insertBefore = nullptr);
+
+extern llvm::CallInst *LLVMCallInst(llvm::Function *func, llvm::Value *arg0, llvm::Value *arg1, llvm::Value *arg2,
+                                    llvm::Value *arg3, const llvm::Twine &name,
+                                    llvm::Instruction *insertBefore = nullptr);
+
+extern llvm::CallInst *LLVMCallInst(llvm::Function *func, llvm::Value *arg0, llvm::Value *arg1, llvm::Value *arg2,
+                                    llvm::Value *arg3, llvm::Value *arg4, const llvm::Twine &name,
+                                    llvm::Instruction *insertBefore = nullptr);
+
+extern llvm::CallInst *LLVMCallInst(llvm::Function *func, llvm::Value *arg0, llvm::Value *arg1, llvm::Value *arg2,
+                                    llvm::Value *arg3, llvm::Value *arg4, llvm::Value *arg5, const llvm::Twine &name,
+                                    llvm::Instruction *insertBefore = nullptr);
+
+extern llvm::GetElementPtrInst *LLVMGEPInst(llvm::Value *ptr, llvm::Type *ptrElType, llvm::Value *offset,
+                                            const char *name, llvm::Instruction *insertBefore);
+
+/** Mask-related helpers */
+
+/** Given an llvm::Value represinting a vector mask, see if the value is a
+    constant.  If so, return true and set *bits to be the integer mask
+    found by taking the high bits of the mask values in turn and
+    concatenating them into a single integer.  In other words, given the
+    4-wide mask: < 0xffffffff, 0, 0, 0xffffffff >, we have 0b1001 = 9.
+ */
+extern bool GetMaskFromValue(llvm::Value *factor, uint64_t *mask);
+
+/** Determines if the given mask value is all on, all off, mixed, or
+    unknown at compile time.
+*/
+extern MaskStatus GetMaskStatusFromValue(llvm::Value *mask, int vecWidth = -1);
+
+/** Add uwtable attribute for function, windows specific.
+ */
+extern void AddUWTableFuncAttr(llvm::Function *fn);
+
+#ifdef ISPC_XE_ENABLED
+/** This is utility function to determine memory in which pointer was created.
+    For now we use only 3 values:
+    Global is used for pointers passed to kernel externally
+    Private is used for pointers created locally with alloca
+    Generic is currently used to identify Global variables
+*/
 extern AddressSpace GetAddressSpace(llvm::Value *v);
+
+#if ISPC_LLVM_VERSION >= ISPC_LLVM_16_0
+/** Fix function attribute by removing input function attr and adding memory effect instead.
+    https://reviews.llvm.org/D135780
+*/
+extern void FixFunctionAttribute(llvm::Function &Fn, llvm::Attribute::AttrKind attr, llvm::MemoryEffects memEf);
+#endif
 #endif
 } // namespace ispc
