@@ -1,34 +1,7 @@
 /*
-  Copyright (c) 2010-2021, Intel Corporation
-  All rights reserved.
+  Copyright (c) 2010-2024, Intel Corporation
 
-  Redistribution and use in source and binary forms, with or without
-  modification, are permitted provided that the following conditions are
-  met:
-
-    * Redistributions of source code must retain the above copyright
-      notice, this list of conditions and the following disclaimer.
-
-    * Redistributions in binary form must reproduce the above copyright
-      notice, this list of conditions and the following disclaimer in the
-      documentation and/or other materials provided with the distribution.
-
-    * Neither the name of Intel Corporation nor the names of its
-      contributors may be used to endorse or promote products derived from
-      this software without specific prior written permission.
-
-
-   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
-   IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
-   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
-   PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
-   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
-   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
-   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
-   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
-   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
-   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
-   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+  SPDX-License-Identifier: BSD-3-Clause
 */
 
 /** @file util.cpp
@@ -66,6 +39,25 @@
 #include <set>
 
 #include <llvm/IR/DataLayout.h>
+#include <llvm/Support/FileSystem.h>
+#include <llvm/Support/Path.h>
+
+#ifdef ISPC_HOST_IS_APPLE
+#if ISPC_LLVM_VERSION >= ISPC_LLVM_17_0
+// Provide own definition of std::__libcpp_verbose_abort to avoid missing
+// symbols error in macOS builds.
+// See https://libcxx.llvm.org/UsingLibcxx.html#overriding-the-default-termination-handler
+// It is not quite clear why and where this symbol is used.
+void std::__libcpp_verbose_abort(char const *format, ...) {
+    va_list list;
+    va_start(list, format);
+    vfprintf(stderr, format, list);
+    va_end(list);
+
+    abort();
+}
+#endif // ISPV_LLVM_17_0
+#endif // ISPC_HOST_IS_APPLE
 
 using namespace ispc;
 
@@ -81,7 +73,7 @@ int ispc::TerminalWidth() {
 
 #if defined(ISPC_HOST_IS_WINDOWS)
     HANDLE h = GetStdHandle(STD_OUTPUT_HANDLE);
-    if (h == INVALID_HANDLE_VALUE || h == NULL)
+    if (h == INVALID_HANDLE_VALUE || h == nullptr)
         return 80;
     CONSOLE_SCREEN_BUFFER_INFO bufferInfo = {{0}};
     GetConsoleScreenBufferInfo(h, &bufferInfo);
@@ -95,7 +87,7 @@ int ispc::TerminalWidth() {
 }
 
 static bool lHaveANSIColors() {
-    static bool r = (getenv("TERM") != NULL && strcmp(getenv("TERM"), "dumb") != 0);
+    static bool r = (getenv("TERM") != nullptr && strcmp(getenv("TERM"), "dumb") != 0);
 #ifndef ISPC_HOST_IS_WINDOWS
     r &= (bool)isatty(2);
     r |= g->forceColoredOutput;
@@ -278,8 +270,8 @@ void ispc::PrintWithWordBreaks(const char *buf, int indent, int columnWidth, FIL
 #ifdef ISPC_HOST_IS_WINDOWS
 // we cover for the lack vasprintf and asprintf on windows (also covers mingw)
 int vasprintf(char **sptr, const char *fmt, va_list argv) {
-    int wanted = vsnprintf(*sptr = NULL, 0, fmt, argv);
-    if ((wanted < 0) || ((*sptr = (char *)malloc(1 + wanted)) == NULL))
+    int wanted = vsnprintf(*sptr = nullptr, 0, fmt, argv);
+    if ((wanted < 0) || ((*sptr = (char *)malloc(1 + wanted)) == nullptr))
         return -1;
 
     return vsprintf(*sptr, fmt, argv);
@@ -351,7 +343,7 @@ static void lPrint(const char *type, bool isError, SourcePos p, const char *fmt,
 }
 
 void ispc::Error(SourcePos p, const char *fmt, ...) {
-    if (m != NULL) {
+    if (m != nullptr) {
         ++m->errorCount;
         if ((g->errorLimit != -1) && (g->errorLimit <= m->errorCount - 1))
             return;
@@ -366,7 +358,6 @@ void ispc::Error(SourcePos p, const char *fmt, ...) {
 }
 
 void ispc::Debug(SourcePos p, const char *fmt, ...) {
-#ifndef ISPC_NO_DUMPS
     if (!g->debugPrint || g->quiet)
         return;
 
@@ -374,7 +365,6 @@ void ispc::Debug(SourcePos p, const char *fmt, ...) {
     va_start(args, fmt);
     lPrint("Debug", false, p, fmt, args);
     va_end(args);
-#endif
 }
 
 void ispc::Warning(SourcePos p, const char *fmt, ...) {
@@ -384,7 +374,7 @@ void ispc::Warning(SourcePos p, const char *fmt, ...) {
     if ((turnOffWarnings_it != g->turnOffWarnings.end()) && (turnOffWarnings_it->second == false))
         return;
 
-    if (g->warningsAsErrors && m != NULL)
+    if (g->warningsAsErrors && m != nullptr)
         ++m->errorCount;
 
     if (g->disableWarnings || g->quiet)
@@ -410,7 +400,7 @@ void ispc::PerformanceWarning(SourcePos p, const char *fmt, ...) {
     if (turnOffWarnings_it != g->turnOffWarnings.end())
         return;
 
-    if (g->warningsAsErrors && m != NULL)
+    if (g->warningsAsErrors && m != nullptr)
         ++m->errorCount;
 
     va_list args;
@@ -515,39 +505,23 @@ std::vector<std::string> ispc::MatchStrings(const std::string &str, const std::v
     return std::vector<std::string>();
 }
 
-void ispc::GetDirectoryAndFileName(const std::string &currentDirectory, const std::string &relativeName,
-                                   std::string *directory, std::string *filename) {
-#ifdef ISPC_HOST_IS_WINDOWS
-    char path[MAX_PATH];
-    const char *combPath = PathCombine(path, currentDirectory.c_str(), relativeName.c_str());
-    Assert(combPath != NULL);
-    const char *filenamePtr = PathFindFileName(combPath);
-    *filename = filenamePtr;
-    *directory = std::string(combPath, filenamePtr - combPath);
-#else
-    // We need a fully qualified path.  First, see if the current file name
-    // is fully qualified itself--in that case, the current working
-    // directory isn't needed.
-    // @todo This probably needs to be smarter for Windows...
-    std::string fullPath;
-    if (relativeName[0] == '/')
+std::pair<std::string, std::string> ispc::GetDirectoryAndFileName(const std::string &currentDirectory,
+                                                                  const std::string &relativeName) {
+    llvm::SmallString<256> fullPath;
+
+    if (llvm::sys::path::is_absolute(relativeName)) {
+        // We may actually get not relative path but absolute, then just use it.
         fullPath = relativeName;
-    else {
-        fullPath = g->currentDirectory;
-        if (fullPath[fullPath.size() - 1] != '/')
-            fullPath.push_back('/');
-        fullPath += relativeName;
+    } else {
+        fullPath = currentDirectory;
+        llvm::sys::path::append(fullPath, relativeName);
     }
 
-    // now, we need to separate it into the base name and the directory
-    const char *fp = fullPath.c_str();
-    const char *basenameStart = strrchr(fp, '/');
-    Assert(basenameStart != NULL);
-    ++basenameStart;
-    Assert(basenameStart[0] != '\0');
-    *filename = basenameStart;
-    *directory = std::string(fp, basenameStart - fp);
-#endif // ISPC_HOST_IS_WINDOWS
+    // Extract the directory and file name from the full path
+    std::string filename = llvm::sys::path::filename(fullPath).str();
+    std::string dirname = llvm::sys::path::parent_path(fullPath).str();
+
+    return std::make_pair(dirname, filename);
 }
 
 static std::set<std::string> lGetStringArray(const std::string &str) {
